"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.waitAdfLoaded = exports.scanFromAdf = exports.saveScan = exports.getScanHeight = exports.getScanWidth = void 0;
const HPApi_1 = __importDefault(require("./HPApi"));
const promises_1 = __importDefault(require("fs/promises"));
const JpegUtil_1 = __importDefault(require("./JpegUtil"));
const listening_1 = require("./listening");
const ScanJobSettings_1 = __importDefault(require("./ScanJobSettings"));
const ScanContent_1 = require("./ScanContent");
const delay_1 = require("./delay");
const PathHelper_1 = __importDefault(require("./PathHelper"));
const InputSource_1 = require("./InputSource");
async function waitDeviceUntilItIsReadyToUploadOrCompleted(jobUrl) {
    let job = null;
    let isReadyToUpload = false;
    do {
        job = await HPApi_1.default.getJob(jobUrl);
        if (job.jobState === "Canceled") {
            return job;
        }
        else if (job.pageState === "ReadyToUpload" ||
            job.jobState === "Completed") {
            isReadyToUpload = true;
        }
        else if (job.jobState == "Processing") {
            isReadyToUpload = false;
        }
        else {
            console.log(`Unknown jobState: ${job.jobState}`);
        }
        await (0, delay_1.delay)(300);
    } while (!isReadyToUpload);
    return job;
}
async function TryGetDestination(event) {
    //this code can in some cases be executed before the user actually chooses between Document or Photo
    //so lets fetch the contentType (Document or Photo) until we get a value
    let destination = null;
    for (let i = 0; i < 20; i++) {
        const destinationURI = event.destinationURI;
        if (destinationURI) {
            destination = await HPApi_1.default.getDestination(destinationURI);
            const shortcut = destination.shortcut;
            if (shortcut != null) {
                return destination;
            }
        }
        else {
            console.log(`No destination URI found`);
        }
        console.log(`No shortcut yet available, attempt: ${i + 1}/20`);
        await new Promise((resolve) => setTimeout(resolve, 1000)); //wait 1s
    }
    console.log("Failing to detect destination shortcut");
    console.log(JSON.stringify(destination));
    return null;
}
async function scanProcessing(filePath) {
    const buffer = await promises_1.default.readFile(filePath);
    let height = JpegUtil_1.default.fixSizeWithDNL(buffer);
    if (height != null) {
        // rewrite the fixed file
        await promises_1.default.writeFile(filePath, buffer);
        return height;
    }
    return null;
}
function createScanPage(job, currentPageNumber, filePath, sizeFixed) {
    var _a, _b, _c;
    let height = sizeFixed !== null && sizeFixed !== void 0 ? sizeFixed : job.imageHeight;
    return {
        path: filePath,
        pageNumber: currentPageNumber,
        width: (_a = job.imageWidth) !== null && _a !== void 0 ? _a : 0,
        height: height !== null && height !== void 0 ? height : 0,
        xResolution: (_b = job.xResolution) !== null && _b !== void 0 ? _b : 200,
        yResolution: (_c = job.yResolution) !== null && _c !== void 0 ? _c : 200,
    };
}
async function handleProcessingState(job, inputSource, folder, scanCount, currentPageNumber, filePattern, date) {
    if (job.pageState == "ReadyToUpload" &&
        job.binaryURL != null &&
        job.currentPageNumber != null) {
        console.log(`Ready to download page job page ${job.currentPageNumber} at:`, job.binaryURL);
        const destinationFilePath = PathHelper_1.default.getFileForPage(folder, scanCount, currentPageNumber, filePattern, "jpg", date);
        const filePath = await HPApi_1.default.downloadPage(job.binaryURL, destinationFilePath);
        console.log("Page downloaded to:", filePath);
        let sizeFixed = null;
        if (inputSource == InputSource_1.InputSource.Adf) {
            sizeFixed = await scanProcessing(filePath);
            if (sizeFixed == null) {
                console.log(`File size has not been fixed, DNF may not have been found and approximate height is: ${job.imageHeight}`);
            }
        }
        return createScanPage(job, currentPageNumber, filePath, sizeFixed);
    }
    else {
        console.log(`Unknown pageState: ${job.pageState}`);
        await (0, delay_1.delay)(200);
        return null;
    }
}
async function executeScanJob(scanJobSettings, inputSource, folder, scanCount, scanJobContent, filePattern) {
    const jobUrl = await HPApi_1.default.postJob(scanJobSettings);
    console.log("New job created:", jobUrl);
    let job = await HPApi_1.default.getJob(jobUrl);
    while (job.jobState !== "Completed") {
        job = await waitDeviceUntilItIsReadyToUploadOrCompleted(jobUrl);
        if (job.jobState == "Completed") {
            continue;
        }
        if (job.jobState === "Processing") {
            const page = await handleProcessingState(job, inputSource, folder, scanCount, scanJobContent.elements.length + 1, filePattern, new Date());
            job = await HPApi_1.default.getJob(jobUrl);
            if (page != null && job.jobState != "Canceled") {
                scanJobContent.elements.push(page);
            }
        }
        else if (job.jobState === "Canceled") {
            console.log("Job cancelled by device");
            break;
        }
        else {
            console.log(`Unhandled jobState: ${job.jobState}`);
            await (0, delay_1.delay)(200);
        }
    }
    console.log(`Job state: ${job.jobState}, totalPages: ${scanJobContent.elements.length}:`);
    return job.jobState;
}
async function waitScanNewPageRequest(compEventURI) {
    let startNewScanJob = false;
    let wait = true;
    while (wait) {
        await new Promise((resolve) => setTimeout(resolve, 1000)); //wait 1s
        let walkupScanToCompEvent = await HPApi_1.default.getWalkupScanToCompEvent(compEventURI);
        let message = walkupScanToCompEvent.eventType;
        if (message === "ScanNewPageRequested") {
            startNewScanJob = true;
            wait = false;
        }
        else if (message === "ScanPagesComplete") {
            wait = false;
        }
        else if (message === "ScanRequested") {
            // continue waiting
        }
        else {
            wait = false;
            console.log(`Unknown eventType: ${message}`);
        }
    }
    return startNewScanJob;
}
async function executeScanJobs(scanJobSettings, inputSource, folder, scanCount, scanJobContent, firstEvent, deviceCapabilities, filePattern) {
    let jobState = await executeScanJob(scanJobSettings, inputSource, folder, scanCount, scanJobContent, filePattern);
    let lastEvent = firstEvent;
    if (jobState === "Completed" &&
        lastEvent.compEventURI &&
        inputSource !== InputSource_1.InputSource.Adf &&
        lastEvent.destinationURI &&
        deviceCapabilities.supportsMultiItemScanFromPlaten) {
        lastEvent = await (0, listening_1.waitForScanEvent)(lastEvent.destinationURI, lastEvent.agingStamp);
        if (!lastEvent.compEventURI) {
            return;
        }
        let startNewScanJob = await waitScanNewPageRequest(lastEvent.compEventURI);
        while (startNewScanJob) {
            jobState = await executeScanJob(scanJobSettings, inputSource, folder, scanCount, scanJobContent, filePattern);
            if (jobState !== "Completed") {
                return;
            }
            if (!lastEvent.destinationURI) {
                break;
            }
            lastEvent = await (0, listening_1.waitForScanEvent)(lastEvent.destinationURI, lastEvent.agingStamp);
            if (!lastEvent.compEventURI) {
                return;
            }
            startNewScanJob = await waitScanNewPageRequest(lastEvent.compEventURI);
        }
    }
}
async function mergeToPdf(folder, scanCount, scanJobContent, filePattern, date) {
    if (scanJobContent.elements.length > 0) {
        const pdfFilePath = PathHelper_1.default.getFileForScan(folder, scanCount, filePattern, "pdf", date);
        await (0, ScanContent_1.createPdfFrom)(scanJobContent, pdfFilePath);
        scanJobContent.elements.forEach((e) => promises_1.default.unlink(e.path));
        return pdfFilePath;
    }
    console.log(`No page available to build a pdf file`);
    return null;
}
function displayPdfScan(pdfFilePath, scanJobContent) {
    if (pdfFilePath === null) {
        console.log(`Pdf generated has not been generated`);
        return;
    }
    console.log(`The following page(s) have been rendered inside '${pdfFilePath}': `);
    scanJobContent.elements.forEach((e) => console.log(`\t- page ${e.pageNumber.toString().padStart(3, " ")} - ${e.width}x${e.height}`));
}
function displayJpegScan(scanJobContent) {
    scanJobContent.elements.forEach((e) => console.log(`\t- page ${e.pageNumber.toString().padStart(3, " ")} - ${e.width}x${e.height} - ${e.path}`));
}
function isPdf(destination) {
    if (destination.shortcut === "SavePDF" ||
        destination.shortcut === "EmailPDF" ||
        destination.shortcut == "SaveDocument1") {
        return true;
    }
    else if (destination.shortcut === "SaveJPEG" ||
        destination.shortcut === "SavePhoto1") {
        return false;
    }
    else {
        console.log(`Unexpected shortcut received: ${destination.shortcut}, considering it as non pdf target!`);
        return false;
    }
}
function getScanWidth(scanConfig, inputSource, deviceCapabilities) {
    if (scanConfig.width && scanConfig.width > 0) {
        const maxWidth = inputSource === InputSource_1.InputSource.Adf
            ? deviceCapabilities.adfMaxWidth
            : deviceCapabilities.platenMaxWidth;
        if (maxWidth && scanConfig.width > maxWidth) {
            return maxWidth;
        }
        else {
            return scanConfig.width;
        }
    }
    else {
        return null;
    }
}
exports.getScanWidth = getScanWidth;
function getScanHeight(scanConfig, inputSource, deviceCapabilities) {
    if (scanConfig.height && scanConfig.height > 0) {
        const maxHeight = inputSource === InputSource_1.InputSource.Adf
            ? deviceCapabilities.adfMaxHeight
            : deviceCapabilities.platenMaxHeight;
        if (maxHeight && scanConfig.height > maxHeight) {
            return maxHeight;
        }
        else {
            return scanConfig.height;
        }
    }
    else {
        return null;
    }
}
exports.getScanHeight = getScanHeight;
async function saveScan(event, folder, tempFolder, scanCount, deviceCapabilities, scanConfig) {
    if (event.compEventURI) {
        const proceedToScan = await (0, listening_1.waitScanRequest)(event.compEventURI);
        if (!proceedToScan) {
            return;
        }
    }
    const destination = await TryGetDestination(event);
    if (!destination) {
        console.log("No shortcut selected!");
        return;
    }
    console.log("Selected shortcut: " + destination.shortcut);
    let toPdf;
    let destinationFolder;
    let contentType;
    if (isPdf(destination)) {
        toPdf = true;
        contentType = "Document";
        destinationFolder = tempFolder;
        console.log(`Scan will be converted to pdf, using ${destinationFolder} as temp scan output directory for individual pages`);
    }
    else {
        toPdf = false;
        contentType = "Photo";
        destinationFolder = folder;
    }
    const isDuplex = destination.scanPlexMode != null && destination.scanPlexMode != "Simplex";
    console.log("ScanPlexMode is : " + destination.scanPlexMode);
    const scanStatus = await HPApi_1.default.getScanStatus();
    console.log("Afd is : " + scanStatus.adfState);
    const inputSource = scanStatus.getInputSource();
    const scanWidth = getScanWidth(scanConfig, inputSource, deviceCapabilities);
    const scanHeight = getScanHeight(scanConfig, inputSource, deviceCapabilities);
    const scanJobSettings = new ScanJobSettings_1.default(inputSource, contentType, scanConfig.resolution, scanWidth, scanHeight, isDuplex);
    const scanJobContent = { elements: [] };
    const scanDate = new Date();
    await executeScanJobs(scanJobSettings, inputSource, destinationFolder, scanCount, scanJobContent, event, deviceCapabilities, scanConfig.directoryConfig.filePattern);
    console.log(`Scan of page(s) completed totalPages: ${scanJobContent.elements.length}:`);
    if (toPdf) {
        const pdfFilePath = await mergeToPdf(folder, scanCount, scanJobContent, scanConfig.directoryConfig.filePattern, scanDate);
        displayPdfScan(pdfFilePath, scanJobContent);
    }
    else {
        displayJpegScan(scanJobContent);
    }
}
exports.saveScan = saveScan;
async function scanFromAdf(scanCount, folder, tempFolder, adfAutoScanConfig, deviceCapabilities, date) {
    let destinationFolder;
    let contentType;
    if (adfAutoScanConfig.generatePdf) {
        contentType = "Document";
        destinationFolder = tempFolder;
        console.log(`Scan will be converted to pdf, using ${destinationFolder} as temp scan output directory for individual pages`);
    }
    else {
        contentType = "Photo";
        destinationFolder = folder;
    }
    const scanWidth = getScanWidth(adfAutoScanConfig, InputSource_1.InputSource.Adf, deviceCapabilities);
    const scanHeight = getScanHeight(adfAutoScanConfig, InputSource_1.InputSource.Adf, deviceCapabilities);
    const scanJobSettings = new ScanJobSettings_1.default(InputSource_1.InputSource.Adf, contentType, adfAutoScanConfig.resolution, scanWidth, scanHeight, adfAutoScanConfig.isDuplex);
    const scanJobContent = { elements: [] };
    await executeScanJob(scanJobSettings, InputSource_1.InputSource.Adf, destinationFolder, scanCount, scanJobContent, adfAutoScanConfig.directoryConfig.filePattern);
    console.log(`Scan of page(s) completed, total pages: ${scanJobContent.elements.length}:`);
    if (adfAutoScanConfig.generatePdf) {
        const pdfFilePath = await mergeToPdf(folder, scanCount, scanJobContent, adfAutoScanConfig.directoryConfig.filePattern, date);
        displayPdfScan(pdfFilePath, scanJobContent);
    }
    else {
        displayJpegScan(scanJobContent);
    }
}
exports.scanFromAdf = scanFromAdf;
async function waitAdfLoaded(pollingInterval, startScanDelay) {
    let ready = false;
    while (!ready) {
        let scanStatus = await HPApi_1.default.getScanStatus();
        while (!scanStatus.isLoaded()) {
            await (0, delay_1.delay)(pollingInterval);
            scanStatus = await HPApi_1.default.getScanStatus();
        }
        console.log(`ADF load detected`);
        let loaded = true;
        let counter = 0;
        const shortPollingInterval = 500;
        while (loaded && counter < startScanDelay) {
            await (0, delay_1.delay)(shortPollingInterval);
            scanStatus = await HPApi_1.default.getScanStatus();
            loaded = scanStatus.isLoaded();
            counter += shortPollingInterval;
        }
        if (loaded && counter >= startScanDelay) {
            ready = true;
            console.log(`ADF still loaded, proceeding`);
        }
        else {
            console.log(`ADF not loaded anymore, waiting...`);
        }
    }
}
exports.waitAdfLoaded = waitAdfLoaded;
//# sourceMappingURL=scanProcessing.js.map